package router

import (
	"context"
	"errors"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/redistool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/testing/mock_redis"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/testing/testhelpers"
	"go.uber.org/mock/gomock"
	clocktesting "k8s.io/utils/clock/testing"
)

var (
	_ Registerer = &RedisTracker{}
	_ Tracker    = &RedisTracker{}
	_ Querier    = &RedisTracker{}
)

const (
	legacySelfURL = "grpc://1.1.1.1:10"
	selfURL       = "grpc:1.1.1.1|10"
	ttl           = time.Minute
)

func TestRegisterConnection(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, tm := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl),
		bl.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()).
			Do(func(keys []int64, ttl time.Duration, kvs ...redistool.BuilderKV[string]) {
				require.Len(t, kvs, 1)
				assert.Equal(t, legacySelfURL, kvs[0].HashKey)
				assert.Equal(t, tm.Add(ttl).Unix(), kvs[0].Value.ExpiresAt)
			}),
		bl.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b),
		b.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()).
			Do(func(keys []int64, ttl time.Duration, kvs ...redistool.BuilderKV[string]) {
				require.Len(t, kvs, 1)
				assert.Equal(t, selfURL, kvs[0].HashKey)
				assert.Equal(t, tm.Add(ttl).Unix(), kvs[0].Value.ExpiresAt)
			}),
		b.EXPECT().
			Do(gomock.Any()),
	)
	bld := r.RegistrationBuilder()
	bld.Register(ttl, testhelpers.AgentID)
	assert.NoError(t, bld.Do(context.Background()))
}

func TestUnregisterConnection(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl1 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	bl2 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b1 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b2 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, _ := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl1),
		bl1.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		bl1.EXPECT().
			Do(gomock.Any()),
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl2),
		bl2.EXPECT().
			Unset([]int64{testhelpers.AgentID}, legacySelfURL),
		bl2.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b1),
		b1.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		b1.EXPECT().
			Do(gomock.Any()),
		hash.EXPECT().
			IOBuilder().
			Return(b2),
		b2.EXPECT().
			Unset([]int64{testhelpers.AgentID}, selfURL),
		b2.EXPECT().
			Do(gomock.Any()),
	)

	bld := r.RegistrationBuilder()
	bld.Register(ttl, testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))

	bld = r.RegistrationBuilder()
	bld.Unregister(testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))
}

func TestUnregisterConnection_Batched(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, _ := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl),
		bl.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		bl.EXPECT().
			Unset([]int64{testhelpers.AgentID}, legacySelfURL),
		bl.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b),
		b.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		b.EXPECT().
			Unset([]int64{testhelpers.AgentID}, selfURL),
		b.EXPECT().
			Do(gomock.Any()),
	)

	bld := r.RegistrationBuilder()
	bld.Register(ttl, testhelpers.AgentID)
	bld.Unregister(testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))
}

func TestUnregisterConnection_TwoConnections(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl1 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	bl2 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	bl3 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	bl4 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b1 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b2 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b3 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b4 := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, _ := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl1),
		bl1.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		bl1.EXPECT().
			Do(gomock.Any()),
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl2),
		bl2.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		bl2.EXPECT().
			Do(gomock.Any()),
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl3),
		bl3.EXPECT().
			Unset([]int64{testhelpers.AgentID}, legacySelfURL),
		bl3.EXPECT().
			Do(gomock.Any()),
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl4),
		bl4.EXPECT().
			Unset([]int64{testhelpers.AgentID}, legacySelfURL),
		bl4.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b1),
		b1.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		b1.EXPECT().
			Do(gomock.Any()),
		hash.EXPECT().
			IOBuilder().
			Return(b2),
		b2.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()),
		b2.EXPECT().
			Do(gomock.Any()),
		hash.EXPECT().
			IOBuilder().
			Return(b3),
		b3.EXPECT().
			Unset([]int64{testhelpers.AgentID}, selfURL),
		b3.EXPECT().
			Do(gomock.Any()),
		hash.EXPECT().
			IOBuilder().
			Return(b4),
		b4.EXPECT().
			Unset([]int64{testhelpers.AgentID}, selfURL),
		b4.EXPECT().
			Do(gomock.Any()),
	)

	bld := r.RegistrationBuilder()
	bld.Register(ttl, testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))

	bld = r.RegistrationBuilder()
	bld.Register(ttl, testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))

	bld = r.RegistrationBuilder()
	bld.Unregister(testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))

	bld = r.RegistrationBuilder()
	bld.Unregister(testhelpers.AgentID)
	require.NoError(t, bld.Do(context.Background()))
}

func TestKASURLsByAgentID_HappyPath(t *testing.T) {
	r, hash, legacyHash, _ := setupTracker(t)
	legacyHash.EXPECT().
		Scan(gomock.Any(), testhelpers.AgentID, gomock.Any()).
		Do(func(ctx context.Context, key int64, cb redistool.ScanCallback) error {
			done, err := cb(legacySelfURL, nil, nil)
			require.NoError(t, err)
			assert.False(t, done)
			return nil
		})
	hash.EXPECT().
		Scan(gomock.Any(), testhelpers.AgentID, gomock.Any()).
		Do(func(ctx context.Context, key int64, cb redistool.ScanCallback) error {
			done, err := cb(selfURL, nil, nil)
			require.NoError(t, err)
			assert.False(t, done)
			return nil
		})
	kasURLs, err := r.KASURLsByAgentID(context.Background(), testhelpers.AgentID)
	require.NoError(t, err)
	assert.Equal(t, []string{selfURL}, kasURLs)
}

func TestKASURLsByAgentID_ScanError(t *testing.T) {
	r, hash, legacyHash, _ := setupTracker(t)
	legacyHash.EXPECT().
		Scan(gomock.Any(), testhelpers.AgentID, gomock.Any()).
		Do(func(ctx context.Context, key int64, cb redistool.ScanCallback) error {
			done, err := cb("", nil, errors.New("intended error1"))
			require.NoError(t, err)
			assert.False(t, done)
			return nil
		})
	hash.EXPECT().
		Scan(gomock.Any(), testhelpers.AgentID, gomock.Any()).
		Do(func(ctx context.Context, key int64, cb redistool.ScanCallback) error {
			done, err := cb("", nil, errors.New("intended error2"))
			require.NoError(t, err)
			assert.False(t, done)
			return nil
		})
	kasURLs, err := r.KASURLsByAgentID(context.Background(), testhelpers.AgentID)
	assert.EqualError(t, err, "intended error1\nintended error2")
	assert.Empty(t, kasURLs)
}

func TestRefresh_NoAgents(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, _ := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl),
		bl.EXPECT().
			Set(gomock.Len(0), ttl, gomock.Any()),
		bl.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b),
		b.EXPECT().
			Set(gomock.Len(0), ttl, gomock.Any()),
		b.EXPECT().
			Do(gomock.Any()),
	)
	builder := r.RegistrationBuilder()
	builder.Refresh(ttl)
	assert.NoError(t, builder.Do(context.Background()))
}

func TestRefresh_OneAgent(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, tm := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl),
		bl.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()).
			Do(func(keys []int64, ttl time.Duration, kvs ...redistool.BuilderKV[string]) {
				require.Len(t, kvs, 1)
				assert.Equal(t, legacySelfURL, kvs[0].HashKey)
				assert.Equal(t, tm.Add(ttl).Unix(), kvs[0].Value.ExpiresAt)
			}),
		bl.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b),
		b.EXPECT().
			Set([]int64{testhelpers.AgentID}, ttl, gomock.Any()).
			Do(func(keys []int64, ttl time.Duration, kvs ...redistool.BuilderKV[string]) {
				require.Len(t, kvs, 1)
				assert.Equal(t, selfURL, kvs[0].HashKey)
				assert.Equal(t, tm.Add(ttl).Unix(), kvs[0].Value.ExpiresAt)
			}),
		b.EXPECT().
			Do(gomock.Any()),
	)
	builder := r.RegistrationBuilder()
	builder.Refresh(ttl, testhelpers.AgentID)
	assert.NoError(t, builder.Do(context.Background()))
}

func TestRefresh_TwoAgents(t *testing.T) {
	ctrl := gomock.NewController(t)
	bl := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	b := mock_redis.NewMockIOBuilder[int64, string](ctrl)
	r, hash, legacyHash, tm := setupTracker(t)
	gomock.InOrder(
		legacyHash.EXPECT().
			IOBuilder().
			Return(bl),
		bl.EXPECT().
			Set([]int64{testhelpers.AgentID, testhelpers.AgentID + 1}, ttl, gomock.Any()).
			Do(func(keys []int64, ttl time.Duration, kvs ...redistool.BuilderKV[string]) {
				require.Len(t, kvs, 1)
				assert.Equal(t, legacySelfURL, kvs[0].HashKey)
				assert.Equal(t, tm.Add(ttl).Unix(), kvs[0].Value.ExpiresAt)
			}),
		bl.EXPECT().
			Do(gomock.Any()),
	)
	gomock.InOrder(
		hash.EXPECT().
			IOBuilder().
			Return(b),
		b.EXPECT().
			Set([]int64{testhelpers.AgentID, testhelpers.AgentID + 1}, ttl, gomock.Any()).
			Do(func(keys []int64, ttl time.Duration, kvs ...redistool.BuilderKV[string]) {
				require.Len(t, kvs, 1)
				assert.Equal(t, selfURL, kvs[0].HashKey)
				assert.Equal(t, tm.Add(ttl).Unix(), kvs[0].Value.ExpiresAt)
			}),
		b.EXPECT().
			Do(gomock.Any()),
	)
	builder := r.RegistrationBuilder()
	builder.Refresh(ttl, testhelpers.AgentID, testhelpers.AgentID+1)
	assert.NoError(t, builder.Do(context.Background()))
}

func setupTracker(t *testing.T) (*RedisTracker, *mock_redis.MockExpiringHashAPI[int64, string], *mock_redis.MockExpiringHashAPI[int64, string], time.Time) {
	ctrl := gomock.NewController(t)
	legacyHash := mock_redis.NewMockExpiringHashAPI[int64, string](ctrl)
	hash := mock_redis.NewMockExpiringHashAPI[int64, string](ctrl)
	tm := time.Now()
	return &RedisTracker{
		legacyOwnPrivateAPIURL: legacySelfURL,
		ownPrivateAPIURL:       selfURL,
		clock:                  clocktesting.NewFakePassiveClock(tm),
		legacyTunnelsByAgentID: legacyHash,
		tunnelsByAgentID:       hash,
	}, hash, legacyHash, tm
}
